"""
'CHROMAT.swatch.swatches'
copyright (c) 2023 hexbenjamin
full license available at /COPYING
"""

from colorsys import rgb_to_hsv, hsv_to_rgb
import random
from typing import (
    Iterable,
    Literal,
    Optional,
    Tuple,
)

from rich.text import Text

from .props import SwatchProp, SwatchTupleProp
from ..utility import SwatchText, get_contrast, parse_hex

# > CONSTANTS <

SCALES = {
    "rgb": [255, 255, 255],
    "hsv": [360, 100, 100],
    "r": 255,
    "g": 255,
    "b": 255,
    "h": 360,
    "s": 100,
    "v": 100,
}


# > CLASSES <
class Swatch:
    """
    A color swatch object.
    """

    def __init__(
        self,
        color: Tuple[int, int, int] = (255, 255, 255),
        mode: Literal["rgb", "hsv"] = "rgb",
        name: Optional[str] = None,
    ) -> None:
        """
        Initialize a Swatch object from a tuple of ints.

        Parameters
        ----------
        color : Tuple[int, int, int], optional
            Input color values as integers
            default (255, 255, 255)
        mode : Literal["rgb", "hsv"], optional
            Mode for interpreting the input color
            default "rgb"
        name : str, optional
            Optional name for the swatch when used in Palettes
            default "Swatch"
        """
        if mode == "rgb":
            self._init_rgb(color)
        elif mode == "hsv":
            self._init_hsv(color)

        self.hex = f"#{self.r.i:02x}{self.g.i:02x}{self.b.i:02x}"
        self.name = name if name is not None else self.hex
        self.relative_luminance = self.get_relative_luminance()

    def _init_rgb(self, color: Tuple[int, int, int]):
        """
        Calculate HSV values from RGB values and assign all properties.

        Parameters
        ----------
        color : tuple[int, int, int]
            RGB tuple to use for initialization.
        """
        self.rgb = SwatchTupleProp(color, "rgb")
        self.r, self.g, self.b = [
            SwatchProp(c, m)
            for c, m in zip(
                color,
                "rgb",
            )
        ]
        self.hsv = SwatchTupleProp(rgb_to_hsv(*self.rgb.f), "hsv")
        self.h, self.s, self.v = [
            SwatchProp(c, m)
            for c, m in zip(
                self.hsv.f,
                "hsv",
            )
        ]

    def _init_hsv(self, color: Tuple[int, int, int]):
        """
        Calculate RGB values from HSV values and assign all properties.

        Parameters
        ----------
        color : tuple[int, int, int]
            HSV tuple to use for initialization.
        """
        self.hsv = SwatchTupleProp(color, "hsv")
        self.h, self.s, self.v = [
            SwatchProp(c, m)
            for c, m in zip(
                color,
                "hsv",
            )
        ]
        self.rgb = SwatchTupleProp(hsv_to_rgb(*self.hsv.f), "rgb")
        self.r, self.g, self.b = [
            SwatchProp(c, m)
            for c, m in zip(
                self.rgb.f,
                "rgb",
            )
        ]

    def get_relative_luminance(self):
        """
        Get the relative luminance from the Swatch's RGB values.
        """
        r, g, b = [
            x / 12.92 if x <= 0.03928 else ((x + 0.055) / 1.055) ** 2.4
            for x in self.rgb.f
        ]

        r *= 0.2126
        g *= 0.7152
        b *= 0.0722

        return r + g + b

    def pick_accent(self) -> "Swatch":
        best = (0, self.__class__())

        for swatch in self.generate_accent_values():
            contrast = get_contrast(self.rel_lum, swatch.rel_lum)

            if contrast > best[0]:
                best = (contrast, swatch)
            if contrast > 4.20:
                self.accent = swatch
                return swatch

        self.accent = best[1]
        return best[1]

    def generate_accent_values(self) -> Iterable["Swatch"]:
        # shuffle range 0-100 for value to test
        values = list(range(0, 101, 2))
        random.shuffle(values)
        for value in values:
            # generate new swatch with 10% swatch's original saturation
            # and a random value from 0-100.
            yield Swatch(
                (
                    self.h.i,
                    self.s.i // 10,
                    value,
                ),
                mode="hsv",
                name=f"{self.name}-accent",
            )

    @classmethod
    def from_hex(cls, hex: str, name: Optional[str] = None) -> "Swatch":
        return cls(parse_hex(hex), mode="rgb", name=name)

    @classmethod
    def from_rgb(
        cls, rgb: Tuple[int, int, int], name: Optional[str] = None
    ) -> "Swatch":
        return cls(rgb, mode="rgb", name=name)

    @classmethod
    def from_hsv(
        cls, hsv: Tuple[int, int, int], name: Optional[str] = None
    ) -> "Swatch":
        return cls(hsv, mode="hsv", name=name)

    @property
    def red(self) -> SwatchProp:
        return self.r

    @property
    def green(self) -> SwatchProp:
        return self.g

    @property
    def blue(self) -> SwatchProp:
        return self.b

    @property
    def hue(self) -> SwatchProp:
        return self.h

    @property
    def saturation(self) -> SwatchProp:
        return self.s

    @property
    def value(self) -> SwatchProp:
        return self.v

    @property
    def rel_lum(self) -> float:
        return self.relative_luminance

    def __repr__(self) -> str:
        return f"Swatch({self.hex})"

    def __str__(self) -> str:
        return self.hex

    def __rich__(self) -> Text:
        return Text.assemble(
            "Swatch('",
            SwatchText(self.hex, self.name),
            "')",
        )



"""
'CHROMAT.utility.console'
copyright (c) 2023 hex benjamin
full license available at /COPYING
"""
from rich.console import Console
from rich.text import Text
from rich.theme import Theme
from rich.traceback import install as rich_traceback_install

chromat_theme = Theme(
    {"body": "#B9B8BA", "heading": "#EEEEEE"},
    inherit=False,
)

CONSOLE = Console(theme=chromat_theme, tab_size=4)

rich_traceback_install()


class SwatchText(Text):
    def __init__(self, swatch_hex: str, swatch_name: str = "Swatch"):
        super().__init__(swatch_name, style=f"bold {swatch_hex}")



from typing import Literal, Tuple, Union


# > CONSTANTS <
SCALES = {
    "r": 255,
    "g": 255,
    "b": 255,
    "h": 360,
    "s": 100,
    "v": 100,
    "rgb": (255, 255, 255),
    "hsv": (360, 100, 100),
}


# > CLASSES <
class SwatchProp:
    """
    Property class for Swatch's R/G/B/H/S/V values.
    """

    def __init__(
        self,
        value: Union[int, float],
        mode: Literal["r", "g", "b", "h", "s", "v"],
    ) -> None:
        """
        Initialize a SwatchProp object from an int or float.

        Parameters
        ----------
        value : Union[int, float]
            Input value. Will auto-convert to the unselected mode.
        mode : Literal["r", "g", "b", "h", "s", "v"]
            Mode to use for the property. This selection determines the
            scaling for int/float conversion.
        """
        self._value = value
        self._scale = SCALES[mode]

    @property
    def integer(self) -> int:
        if isinstance(self._value, float):
            return int(self._value * self._scale)
        return self._value

    @property
    def float(self) -> float:
        if isinstance(self._value, int):
            return self._value / self._scale
        return self._value

    @property
    def i(self):
        return self.integer

    @property
    def f(self):
        return self.float


class SwatchTupleProp:
    """
    Property class for Swatch's RGB/HSV tuples.
    """

    def __init__(
        self,
        value: Union[Tuple[int, int, int], Tuple[float, float, float]],
        mode: Literal["rgb", "hsv"],
    ) -> None:
        """
        Initialize a SwatchProp object from a tuple of ints or floats.

        Parameters
        ----------
        value : Union[Tuple[int, int, int], Tuple[float, float, float]]
            Input values. Will auto-convert to the unselected mode.
        mode : Literal["rgb", "hsv"]
            Mode to use for the property. This selection determines the
            scaling for int/float conversion.
        """
        self._values = value
        self._scales = SCALES[mode]

    def __repr__(self):
        return f"SwatchTupleProp({self.i}, {self.f})"

    @property
    def integer(self) -> Tuple[int, int, int]:
        if all(isinstance(v, float) for v in self._values):
            return tuple(
                [
                    int(c * s)
                    for c, s in zip(
                        self._values,
                        self._scales,
                    )
                ]
            )
        else:
            return tuple([c for c in self._values])

    @property
    def float(self) -> Tuple[float, float, float]:
        if all(isinstance(v, int) for v in self._values):
            return tuple(
                [
                    c / s
                    for c, s in zip(
                        self._values,
                        self._scales,
                    )
                ]
            )
        return tuple(self._values)

    @property
    def i(self):
        return self.integer

    @property
    def f(self):
        return self.float



"""
'CHROMAT.utility.colors'
copyright (c) 2023 hex benjamin
full license available at /COPYING
"""
import re

from typing import Tuple


# > CONSTANTS <
HEX_PATTERN = re.compile(r"^([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$")


# > FUNCTIONS <
def parse_hex(hex_string: str) -> Tuple[int, int, int]:
    """
    Parse a hex color string into a tuple of three integers.
    """
    hex_string = hex_string.strip("#")
    match = HEX_PATTERN.match(hex_string)
    if match is None:
        raise ValueError(
            " ".join(
                [
                    f"Invalid hex color: '{hex_string}'.",
                    "Must be in the format '(#)RRGGBB'.",
                ]
            )
        )
    red, green, blue = [int(c, 16) for c in match.groups()]
    return (red, green, blue)


def get_contrast(
    lum1: float,
    lum2: float,
    min_ratio: float = 4.5,
) -> float:
    """
    Calculate the contrast ratio between two swatches.

    Parameters
    ----------
    swatch1 : Swatch
        First swatch to compare.
    swatch2 : Swatch
        Second swatch to compare.
    min_ratio : float, optional
        Ratio to use for determining contrast.
        default 4.5

    Returns
    -------
    float
        Contrast ratio between the two swatches, 1.0-21.0.
    """
    lums = [lum1, lum2]
    return (max(lums) + 0.05) / (min(lums) + 0.05)



"""
'CHROMAT'
copyright (c) 2023 hex benjamin
full license available at /COPYING
"""

from .utility.console import CONSOLE
from .swatch import Swatch
from .palette import Palette

__all__ = [
    "CONSOLE",
    "utility",
    "Swatch",
    "Palette",
]

LAUNCH_STRING = "CHROMAT v0.0.6"



"""
'CHROMAT.utility'
copyright (c) 2023 hex benjamin
full license available at /COPYING
"""

from .colors import (
    HEX_PATTERN,
    get_contrast,
    parse_hex,
)

from .console import (
    SwatchText,
)

__all__ = [
    "HEX_PATTERN",
    "SwatchText",
    "get_contrast",
    "parse_hex",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chromat', 'chromat.palette', 'chromat.swatch', 'chromat.utility']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.2.0,<3.0.0',
 'numpy>=1.22.2,<2.0.0',
 'replit>=3.2.4,<4.0.0',
 'rich>=13.3.4,<14.0.0',
 'urllib3>=1.26.12,<2.0.0']

setup_kwargs = {
    'name': 'chromat',
    'version': '0.0.6',
    'description': 'color palettes! under heavy construction!',
    'long_description': '\ufeff# chromat: algorithmic color palettes\ncoming soon!\n\nhttps://github.com/hexbenjamin/chromat',
    'author': 'hex benjamin',
    'author_email': 'hex@hexbenjam.in',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/hexbenjamin/chromat',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10.0,<3.11',
}


setup(**setup_kwargs)



"""
'CHROMAT.palette.palettes'
copyright (c) 2023 hexbenjamin
full license available at /COPYING
"""

from typing import Iterable

from ..swatch import Swatch


class Palette:
    def __init__(self, *swatches: Swatch) -> None:
        self.swatches = swatches

    def __iter__(self) -> Iterable[Swatch]:
        return iter(self.swatches)
